@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import contextlib
 import inspect
+import json
 import operator
 import pickle
 import re
@@ -39,7 +40,13 @@ from zarr.core.group import (
 )
 from zarr.core.metadata.v3 import ArrayV3Metadata
 from zarr.core.sync import _collect_aiterator, sync
-from zarr.errors import ContainsArrayError, ContainsGroupError, MetadataValidationError
+from zarr.errors import (
+    ContainsArrayError,
+    ContainsGroupError,
+    MetadataValidationError,
+    ZarrDeprecationWarning,
+    ZarrUserWarning,
+)
 from zarr.storage import LocalStore, MemoryStore, StorePath, ZipStore
 from zarr.storage._common import make_store_path
 from zarr.storage._utils import _join_paths, normalize_path
@@ -52,6 +59,7 @@ if TYPE_CHECKING:
 
     from _pytest.compat import LEGACY_PATH
 
+    from zarr.core.buffer.core import Buffer
     from zarr.core.common import JSON, ZarrFormat
 
 
@@ -203,11 +211,17 @@ def test_group_members(store: Store, zarr_format: ZarrFormat, consolidated_metad
 
     # this warning shows up when extra objects show up in the hierarchy
     warn_context = pytest.warns(
-        UserWarning, match=r"Object at .* is not recognized as a component of a Zarr hierarchy."
+        ZarrUserWarning,
+        match=r"(?:Object at .* is not recognized as a component of a Zarr hierarchy.)|(?:Consolidated metadata is currently not part in the Zarr format 3 specification.)",
     )
     if consolidated_metadata:
-        with warn_context:
-            zarr.consolidate_metadata(store=store, zarr_format=zarr_format)
+        if isinstance(store, ZipStore):
+            with warn_context:
+                with pytest.warns(UserWarning, match="Duplicate name: "):
+                    zarr.consolidate_metadata(store=store, zarr_format=zarr_format)
+        else:
+            with warn_context:
+                zarr.consolidate_metadata(store=store, zarr_format=zarr_format)
         # now that we've consolidated the store, we shouldn't get the warnings from the unrecognized objects anymore
         # we use a nullcontext to handle these cases
         warn_context = contextlib.nullcontext()
@@ -267,7 +281,11 @@ def test_group(store: Store, zarr_format: ZarrFormat) -> None:
     assert dict(bar2.attrs) == {"baz": "qux"}
 
     # update a group's attributes
-    bar2.attrs.update({"name": "bar"})
+    if isinstance(store, ZipStore):
+        with pytest.warns(UserWarning, match="Duplicate name: "):
+            bar2.attrs.update({"name": "bar"})
+    else:
+        bar2.attrs.update({"name": "bar"})
     # bar.attrs was modified in-place
     assert dict(bar2.attrs) == {"baz": "qux", "name": "bar"}
 
@@ -340,7 +358,30 @@ def test_group_getitem(store: Store, zarr_format: ZarrFormat, consolidated: bool
     subsubarray = subgroup.create_array(name="subarray", shape=(10,), chunks=(10,), dtype="uint8")
 
     if consolidated:
-        group = zarr.api.synchronous.consolidate_metadata(store=store, zarr_format=zarr_format)
+        if zarr_format == 3:
+            with pytest.warns(  # noqa: PT031
+                ZarrUserWarning,
+                match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+            ):
+                if isinstance(store, ZipStore):
+                    with pytest.warns(UserWarning, match="Duplicate name: "):
+                        group = zarr.api.synchronous.consolidate_metadata(
+                            store=store, zarr_format=zarr_format
+                        )
+                else:
+                    group = zarr.api.synchronous.consolidate_metadata(
+                        store=store, zarr_format=zarr_format
+                    )
+        else:
+            if isinstance(store, ZipStore):
+                with pytest.warns(UserWarning, match="Duplicate name: "):
+                    group = zarr.api.synchronous.consolidate_metadata(
+                        store=store, zarr_format=zarr_format
+                    )
+            else:
+                group = zarr.api.synchronous.consolidate_metadata(
+                    store=store, zarr_format=zarr_format
+                )
         # we're going to assume that `group.metadata` is correct, and reuse that to focus
         # on indexing in this test. Other tests verify the correctness of group.metadata
         object.__setattr__(
@@ -398,8 +439,11 @@ def test_group_get_with_default(store: Store, zarr_format: ZarrFormat) -> None:
 
     # now with a group
     subgroup = group.require_group("subgroup")
-    subgroup.attrs["foo"] = "bar"
-
+    if isinstance(store, ZipStore):
+        with pytest.warns(UserWarning, match="Duplicate name: "):
+            subgroup.attrs["foo"] = "bar"
+    else:
+        subgroup.attrs["foo"] = "bar"
     result = group.get("subgroup", 8)
     assert result.attrs["foo"] == "bar"
 
@@ -417,7 +461,22 @@ def test_group_delitem(store: Store, zarr_format: ZarrFormat, consolidated: bool
     subarray = group.create_array(name="subarray", shape=(10,), chunks=(10,), dtype="uint8")
 
     if consolidated:
-        group = zarr.api.synchronous.consolidate_metadata(store=store, zarr_format=zarr_format)
+        if zarr_format == 3:
+            with pytest.warns(  # noqa: PT031
+                ZarrUserWarning,
+                match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+            ):
+                if isinstance(store, ZipStore):
+                    with pytest.warns(UserWarning, match="Duplicate name: "):
+                        group = zarr.api.synchronous.consolidate_metadata(
+                            store=store, zarr_format=zarr_format
+                        )
+                else:
+                    group = zarr.api.synchronous.consolidate_metadata(
+                        store=store, zarr_format=zarr_format
+                    )
+        else:
+            group = zarr.api.synchronous.consolidate_metadata(store=store, zarr_format=zarr_format)
         object.__setattr__(
             subgroup.metadata, "consolidated_metadata", ConsolidatedMetadata(metadata={})
         )
@@ -512,7 +571,22 @@ def test_group_child_iterators(store: Store, zarr_format: ZarrFormat, consolidat
     expected_arrays = list(zip(expected_array_keys, expected_array_values, strict=False))
 
     if consolidate:
-        group = zarr.consolidate_metadata(store)
+        if zarr_format == 3:
+            with pytest.warns(  # noqa: PT031
+                ZarrUserWarning,
+                match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+            ):
+                if isinstance(store, ZipStore):
+                    with pytest.warns(UserWarning, match="Duplicate name: "):
+                        group = zarr.consolidate_metadata(store)
+                else:
+                    group = zarr.consolidate_metadata(store)
+        else:
+            if isinstance(store, ZipStore):
+                with pytest.warns(UserWarning, match="Duplicate name: "):
+                    group = zarr.consolidate_metadata(store)
+            else:
+                group = zarr.consolidate_metadata(store)
         if zarr_format == 2:
             metadata = {
                 "subarray": {
@@ -608,7 +682,11 @@ def test_group_update_attributes(store: Store, zarr_format: ZarrFormat) -> None:
     group = Group.from_store(store, zarr_format=zarr_format, attributes=attrs)
     assert group.attrs == attrs
     new_attrs = {"bar": 100}
-    new_group = group.update_attributes(new_attrs)
+    if isinstance(store, ZipStore):
+        with pytest.warns(UserWarning, match="Duplicate name: "):
+            new_group = group.update_attributes(new_attrs)
+    else:
+        new_group = group.update_attributes(new_attrs)
 
     updated_attrs = attrs.copy()
     updated_attrs.update(new_attrs)
@@ -623,7 +701,11 @@ async def test_group_update_attributes_async(store: Store, zarr_format: ZarrForm
     group = Group.from_store(store, zarr_format=zarr_format, attributes=attrs)
     assert group.attrs == attrs
     new_attrs = {"bar": 100}
-    new_group = await group.update_attributes_async(new_attrs)
+    if isinstance(store, ZipStore):
+        with pytest.warns(UserWarning, match="Duplicate name: "):
+            new_group = await group.update_attributes_async(new_attrs)
+    else:
+        new_group = await group.update_attributes_async(new_attrs)
     assert new_group.attrs == new_attrs
 
 
@@ -648,8 +730,12 @@ def test_group_create_array(
         array = group.create_array(name=name, shape=shape, dtype=dtype)
         array[:] = data
     elif method == "array":
-        with pytest.warns(DeprecationWarning, match=r"Group\.create_array instead\."):
-            array = group.array(name=name, data=data, shape=shape, dtype=dtype)
+        with pytest.warns(ZarrDeprecationWarning, match=r"Group\.create_array instead\."):
+            with pytest.warns(
+                ZarrUserWarning,
+                match="The `compressor` argument is deprecated. Use `compressors` instead.",
+            ):
+                array = group.array(name=name, data=data, shape=shape, dtype=dtype)
     else:
         raise AssertionError
 
@@ -660,8 +746,12 @@ def test_group_create_array(
                 a[:] = data
         elif method == "array":
             with pytest.raises(ContainsArrayError):  # noqa: PT012
-                with pytest.warns(DeprecationWarning, match=r"Group\.create_array instead\."):
-                    a = group.array(name=name, shape=shape, dtype=dtype)
+                with pytest.warns(ZarrDeprecationWarning, match=r"Group\.create_array instead\."):
+                    with pytest.warns(
+                        ZarrUserWarning,
+                        match="The `compressor` argument is deprecated. Use `compressors` instead.",
+                    ):
+                        a = group.array(name=name, shape=shape, dtype=dtype)
                 a[:] = data
 
     assert array.path == normalize_path(name)
@@ -1024,7 +1114,11 @@ async def test_asyncgroup_update_attributes(store: Store, zarr_format: ZarrForma
         store=store, zarr_format=zarr_format, attributes=attributes_old
     )
 
-    agroup_new_attributes = await agroup.update_attributes(attributes_new)
+    if isinstance(store, ZipStore):
+        with pytest.warns(UserWarning, match="Duplicate name"):
+            agroup_new_attributes = await agroup.update_attributes(attributes_new)
+    else:
+        agroup_new_attributes = await agroup.update_attributes(attributes_new)
     attributes_updated = attributes_old.copy()
     attributes_updated.update(attributes_new)
     assert agroup_new_attributes.attrs == attributes_updated
@@ -1099,8 +1193,16 @@ async def test_group_members_async(store: Store, consolidated_metadata: bool) ->
     assert all_children == expected
 
     if consolidated_metadata:
-        await zarr.api.asynchronous.consolidate_metadata(store=store)
-        group = await zarr.api.asynchronous.open_group(store=store)
+        with pytest.warns(  # noqa: PT031
+            ZarrUserWarning,
+            match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+        ):
+            if isinstance(store, ZipStore):
+                with pytest.warns(UserWarning, match="Duplicate name"):
+                    await zarr.api.asynchronous.consolidate_metadata(store=store)
+            else:
+                await zarr.api.asynchronous.consolidate_metadata(store=store)
+    group = await zarr.api.asynchronous.open_group(store=store)
 
     nmembers = await group.nmembers(max_depth=None)
     assert nmembers == 6
@@ -1194,27 +1296,27 @@ def test_create_dataset_with_data(store: Store, zarr_format: ZarrFormat) -> None
     """
     root = Group.from_store(store=store, zarr_format=zarr_format)
     arr = np.random.random((5, 5))
-    with pytest.warns(DeprecationWarning, match=r"Group\.create_array instead\."):
+    with pytest.warns(ZarrDeprecationWarning, match=r"Group\.create_array instead\."):
         data = root.create_dataset("random", data=arr, shape=arr.shape)
     np.testing.assert_array_equal(np.asarray(data), arr)
 
 
 async def test_create_dataset(store: Store, zarr_format: ZarrFormat) -> None:
     root = await AsyncGroup.from_store(store=store, zarr_format=zarr_format)
-    with pytest.warns(DeprecationWarning, match=r"Group\.create_array instead\."):
+    with pytest.warns(ZarrDeprecationWarning, match=r"Group\.create_array instead\."):
         foo = await root.create_dataset("foo", shape=(10,), dtype="uint8")
     assert foo.shape == (10,)
 
     with (
         pytest.raises(ContainsArrayError),
-        pytest.warns(DeprecationWarning, match=r"Group\.create_array instead\."),
+        pytest.warns(ZarrDeprecationWarning, match=r"Group\.create_array instead\."),
     ):
         await root.create_dataset("foo", shape=(100,), dtype="int8")
 
     _ = await root.create_group("bar")
     with (
         pytest.raises(ContainsGroupError),
-        pytest.warns(DeprecationWarning, match=r"Group\.create_array instead\."),
+        pytest.warns(ZarrDeprecationWarning, match=r"Group\.create_array instead\."),
     ):
         await root.create_dataset("bar", shape=(100,), dtype="int8")
 
@@ -1253,8 +1355,25 @@ async def test_members_name(store: Store, consolidate: bool, zarr_format: ZarrFo
     b.create_array("array", shape=(1,), dtype="uint8")
 
     if consolidate:
-        group = zarr.api.synchronous.consolidate_metadata(store)
-
+        if isinstance(store, ZipStore):
+            with pytest.warns(UserWarning, match="Duplicate name"):  # noqa: PT031
+                if zarr_format == 3:
+                    with pytest.warns(
+                        ZarrUserWarning,
+                        match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+                    ):
+                        group = zarr.api.synchronous.consolidate_metadata(store)
+                else:
+                    group = zarr.api.synchronous.consolidate_metadata(store)
+        else:
+            if zarr_format == 3:
+                with pytest.warns(
+                    ZarrUserWarning,
+                    match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+                ):
+                    group = zarr.api.synchronous.consolidate_metadata(store)
+            else:
+                group = zarr.api.synchronous.consolidate_metadata(store)
     result = group["a"]["b"]
     assert result.name == "/a/b"
 
@@ -1283,6 +1402,21 @@ def test_open_mutable_mapping_sync():
     assert isinstance(group.store_path.store, MemoryStore)
 
 
+async def test_open_ambiguous_node():
+    zarr_json_bytes = default_buffer_prototype().buffer.from_bytes(
+        json.dumps({"zarr_format": 3, "node_type": "group"}).encode("utf-8")
+    )
+    zgroup_bytes = default_buffer_prototype().buffer.from_bytes(
+        json.dumps({"zarr_format": 2}).encode("utf-8")
+    )
+    store: dict[str, Buffer] = {"zarr.json": zarr_json_bytes, ".zgroup": zgroup_bytes}
+    with pytest.warns(
+        ZarrUserWarning,
+        match=r"Both zarr\.json \(Zarr format 3\) and \.zgroup \(Zarr format 2\) metadata objects exist at",
+    ):
+        await AsyncGroup.open(store, zarr_format=None)
+
+
 class TestConsolidated:
     async def test_group_getitem_consolidated(self, store: Store) -> None:
         root = await AsyncGroup.from_store(store=store)
@@ -1303,7 +1437,15 @@ class TestConsolidated:
         x1 = await x0.create_group("x1")
         await x1.create_group("x2")
 
-        await zarr.api.asynchronous.consolidate_metadata(store)
+        with pytest.warns(  # noqa: PT031
+            ZarrUserWarning,
+            match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+        ):
+            if isinstance(store, ZipStore):
+                with pytest.warns(UserWarning, match="Duplicate name"):
+                    await zarr.api.asynchronous.consolidate_metadata(store)
+            else:
+                await zarr.api.asynchronous.consolidate_metadata(store)
 
         # On disk, we've consolidated all the metadata in the root zarr.json
         group = await zarr.api.asynchronous.open(store=store)
@@ -1360,7 +1502,15 @@ class TestConsolidated:
         x2 = await x1.create_group("x2")
         await x2.create_array("data", shape=(1,), dtype="uint8")
 
-        await zarr.api.asynchronous.consolidate_metadata(store)
+        with pytest.warns(  # noqa: PT031
+            ZarrUserWarning,
+            match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+        ):
+            if isinstance(store, ZipStore):
+                with pytest.warns(UserWarning, match="Duplicate name"):
+                    await zarr.api.asynchronous.consolidate_metadata(store)
+            else:
+                await zarr.api.asynchronous.consolidate_metadata(store)
 
         group = await zarr.api.asynchronous.open_consolidated(store=store)
         assert len(group.metadata.consolidated_metadata.metadata) == 2
@@ -1384,7 +1534,11 @@ class TestConsolidated:
 
         # Now create consolidated metadata...
         root.create_group("g0")
-        zarr.consolidate_metadata(store)
+        with pytest.warns(
+            ZarrUserWarning,
+            match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+        ):
+            zarr.consolidate_metadata(store)
 
         # and explicitly ignore it.
         group = zarr.open_group(store=store, use_consolidated=False)
@@ -1404,7 +1558,11 @@ class TestConsolidated:
 
         # Now create consolidated metadata...
         await root.create_group("g0")
-        await zarr.api.asynchronous.consolidate_metadata(store)
+        with pytest.warns(
+            ZarrUserWarning,
+            match="Consolidated metadata is currently not part in the Zarr format 3 specification.",
+        ):
+            await zarr.api.asynchronous.consolidate_metadata(store)
 
         # and explicitly ignore it.
         group = await zarr.api.asynchronous.open_group(store=store, use_consolidated=False)
